/* gno3dtet: 3D Tetris game for GNOME
 * Copyright (C) 1999, 2000, 2001 Sebastien Nicoud <seb@eseb.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/

#include "common.h"
#include "csound.h"
#include "csettings.h"
#include "cpiece.h"
#include "cgame.h"

const gint CGame::NB_PIECES_MAX = CGAME_NB_PIECES_MAX;
const gint CGame::NB_TRANSFO_MAX = CGAME_NB_TRANSFO_MAX;

gint CGame::nb_pieces = 0;

CPiece *CGame::pieces[CGAME_NB_PIECES_MAX];

const gchar *CGame::piece_descriptors[CGAME_NB_PIECES_MAX] = {
	/* 1 cube */  "1",
	/* 2 cubes */ "1r",
	/* 3 cubes */ "1rr", "1rlu",
	/* 4 cubes */ "1rrr", "1rul", "1rrlu", "1rur", "1udrr",
				  "2rludf", "2udrf", "2rluf",
	/* 5 cubes */ "1rrrr", "2udrru", "2ruru", "3rudrld", "1ruur", "2ruudr",
				  "2rrluu", "1rruu", "1urdr", "1udrrr", "1rurr", "1rudrr",
				  "3furr", "3frur", "3rfdr", "3rudfbr", "2rulf",
				  "3frru", "3frrd", "3frrlu", "3frrld", "3frur",
				  "3frdr", "3rfbdr", "3rfbur", "3frub", "2udfbrr",
	"EOL" // End Of List
};
gint CGame::pieces_set[4][CGAME_NB_PIECES_MAX];
gint CGame::nb_pieces_by_diff[4];

//const gchar *CGame::str_id = "id";
//const gchar *CGame::str_score = "score";
//const gchar *CGame::str_level = "level";
//const gchar *CGame::str_layers = "layers";
//const gchar *CGame::str_timer = "time";
//const gchar *CGame::str_name = "name";
//const gchar *CGame::str_date = "date";
//const gchar *CGame::str_eog = "$";

CGame::CGame()
{
	if (debug) g_print("CGame::CGame()\n");
	_id = 0;
	_name = NULL;
	_date = NULL;
	_timer = 0;
	_level = 0;
	_layers = 0;
	_score = 0;
	_cubes = 0;
	_blocs = 0;
	
	clear_transfos(); // transfos initialized in get_next_piece()
	
//	_pos = 0L;

	_state = state_NEW;
	_settings = g_new(CSettings,1);
	_hw = 0;
	_hh = 0;
 // _world[] initialized in init()
 // _layer[] initialized in init()
	_smooth = TRUE;
	_falling = FALSE;
	_falling_speed = 1;
 //	_piece; initialized in get_next_piece()
	_next_piece = -1;
 //	_roted_piece; initialized in get_next_piece()
 //	_xing_piece; initialized in get_next_piece()
 //	_xed_piece; initialized in get_next_piece()
	_highest_cube = 0;

    for (int i=0; i<CGAME_NB_PIECES_MAX; i++) pieces[i] = NULL;
}

CGame::CGame(const CGame &ind)
{
	if (debug) g_print("CGame::CGame(CGame &)\n");
	free_memory();
	_id = ind._id;
	if (ind._name != NULL) _name = g_strdup(ind._name);
	if (ind._date != NULL) _date = g_strdup(ind._date);
	_timer = ind._timer;
	_level = ind._level;
	_layers = ind._layers;
	_score = ind._score;
	_cubes = ind._cubes;
	_blocs = ind._blocs;

	clear_transfos();
	_nb_transfo = ind._nb_transfo;
	_first_translation = ind._first_translation;
	_last_translation = ind._last_translation;
	_first_rotation = ind._first_rotation;
	_last_rotation = ind._last_rotation;
	_first_transfo = ind._first_transfo;
	for (gint i=0; i<CGAME_NB_TRANSFO_MAX; i++) {
		_transfos[i].fx = ind._transfos[i].fx;
		_transfos[i].t = ind._transfos[i].t;
		_transfos[i].next = ind._transfos[i].next;
		_transfos[i].prec = ind._transfos[i].prec;
	}		
	
//	_pos = ind._pos;

	_state = ind._state;
	_settings = g_new(CSettings,1); *_settings = *(ind._settings);
	_hw = ind._hw;
	_hh = ind._hh;

	gint x,y,z;
	for (x=0; x<MAX_WORLD_WIDTH+2; x++)
		for (y=0; y<MAX_WORLD_HEIGHT+2; y++)
			for (z=0; z<MAX_WORLD_DEPTH+4; z++)
				_world[x][y][z] = ind._world[x][y][z];
	for (z=1; z<MAX_WORLD_DEPTH+4; z++)
		_layer[z] = ind._layer[z];

	_smooth = ind._smooth;
	_falling = ind._falling;
	_falling_speed = ind._falling_speed;
	
	_piece = ind._piece;
	_next_piece = ind._next_piece;
	_roted_piece = ind._roted_piece;
	_xing_piece = ind._xing_piece;
	_xed_piece = ind._xed_piece;

	_highest_cube = ind._highest_cube;
}

CGame::~CGame()
{
	if (debug) g_print("CGame::~CGame\n");
	free_memory();
}

CGame &CGame::operator= (const CGame &ind)
{
	if (debug > 2) g_print("CGame::operator=\n");
	free_memory();
	_id = ind._id;
	if (ind._name != NULL) _name = g_strdup(ind._name);
	if (ind._date != NULL) _date = g_strdup(ind._date);
	_timer = ind._timer;
	_level = ind._level;
	_layers = ind._layers;
	_score = ind._score;
	_cubes = ind._cubes;
	_blocs = ind._blocs;
	
	clear_transfos();
	_nb_transfo = ind._nb_transfo;
	_first_translation = ind._first_translation;
	_last_translation = ind._last_translation;
	_first_rotation = ind._first_rotation;
	_last_rotation = ind._last_rotation;
	_first_transfo = ind._first_transfo;
	for (gint i=0; i<CGAME_NB_TRANSFO_MAX; i++) {
		_transfos[i].fx = ind._transfos[i].fx;
		_transfos[i].t = ind._transfos[i].t;
		_transfos[i].next = ind._transfos[i].next;
		_transfos[i].prec = ind._transfos[i].prec;
	}		
	
//	_pos = ind._pos;

	_state = ind._state;
	_settings = g_new(CSettings,1); *_settings = *(ind._settings);
	_hw = ind._hw;
	_hh = ind._hh;

	gint x,y,z;
	for (x=0; x<MAX_WORLD_WIDTH+2; x++)
		for (y=0; y<MAX_WORLD_HEIGHT+2; y++)
			for (z=0; z<MAX_WORLD_DEPTH+4; z++)
				_world[x][y][z] = ind._world[x][y][z];
	for (z=1; z<MAX_WORLD_DEPTH+4; z++)
		_layer[z] = ind._layer[z];

	_smooth = ind._smooth;
	_falling = ind._falling;
	_falling_speed = ind._falling_speed;
	
	_piece = ind._piece;
	_next_piece = ind._next_piece;
	_roted_piece = ind._roted_piece;
	_xing_piece = ind._xing_piece;
	_xed_piece = ind._xed_piece;
	
	_highest_cube = ind._highest_cube;

	return *this;
}

ostream &operator<< (ostream &os, CGame &ind)
{
//	if (debug > 2) g_print("CGame::operator<<\n");
	os << "game: ";
	os << "  nb of pieces=" << CGame::nb_pieces << endl;
	gint i;
	for (i=0; i<CGame::nb_pieces; i++) {
		os << i << ":"; 
		if (CGame::pieces[i] == NULL) {
			os << "(null)" << endl;
		} else {
			os << *(CGame::pieces[i]);
	}	}
	return os;
}

gboolean CGame::add_transfo(gint transfo, gboolean check_validity)
{
	if (debug > 1) g_print("CGame::add_transfo (%d)\n",transfo);
	
	if (transfo >= trans_LAST) {
		if (debug) g_print("** ERROR **:CGame::add_transfo: invalid transfo: %d\n",transfo);
		return TRUE; // invalid transfo
	}

	if (_nb_transfo+1 >= NB_TRANSFO_MAX) {
		// too many transfos in queue
		if (debug)
			g_print("  max number of transfos reached, action cancelled\n");
		return TRUE;
	}
	if (check_validity)
		// perform only if the transfo is valid
		if (!is_valid_transfo(transfo)) {
//			if (_settings->_sound_support)
//				if (transfo != trans_FALL)
//					csound_play_for_event(CSound::snd_NOMOVE);
			return FALSE; // can't do it
	}
		
	// sound
	if (_settings->_sound_support)
		if (transfo != trans_FALL)
			csound_play_for_event(CSound::snd_MOVE);
	
	_nb_transfo++;
	gboolean transl_or_not_rotation = TRUE; // it's a translation
	if (transfo == trans_XP || transfo == trans_XN
	 || transfo == trans_YP || transfo == trans_YN
	 || transfo == trans_ZP || transfo == trans_ZN)
		transl_or_not_rotation = FALSE; // no, it's actually a rotation

	gint this_transfo = 0;
	gint i;
	for (i=0; i<NB_TRANSFO_MAX; i++)
		if (_transfos[i].fx == trans_NONE) {
			this_transfo = i;
			break; // exit for loop
	}

	if (_nb_transfo <= 1) _first_transfo = this_transfo;
	
	_transfos[this_transfo].fx = transfo;
	if (transfo == trans_FALL) _falling = TRUE;
	else // cancel fall if piece is moved - i.e. transfo other than FALL
		_falling_speed = CGAME_FALLING_SPEED_SLOW;
	_transfos[this_transfo].t = 0;

	if (transl_or_not_rotation) {
		// translation
		if (_last_translation != -1) {
			//// there's a translation already ////
			// prec transfo
			_transfos[_last_translation].next = this_transfo;
			// next transfo
			// there isn't any next transfo
			// prec link
			_transfos[this_transfo].prec = _last_translation;
			// next link
			_transfos[this_transfo].next = -1;
			// first
			// _first_translation unchanged 
			// last
			_last_translation = this_transfo;
		} else {
			///// first translation ////
			// prec transfo
			if (_last_rotation != -1)
				_transfos[_last_rotation].next = this_transfo;
			// next transfo
			// there isn't any next transfo
			// prec link
			_transfos[this_transfo].prec = _last_rotation;
			// next link
			_transfos[this_transfo].next = -1;
			// first
			_first_translation = this_transfo;
			// last
			_last_translation = this_transfo;
		}
	} else {
		// rotation
		if (_last_rotation != -1) {
			//// there's a rotation already ////
			// prec transfo
			_transfos[_last_rotation].next = this_transfo;
			// next transfo
			if (_first_translation != -1)
				_transfos[_first_translation].prec = this_transfo;			
			// prec link
			_transfos[this_transfo].prec = _last_rotation;
			// next link
			_transfos[this_transfo].next = _first_translation;
			// first
			// _first_translation unchanged 
			// last
			_last_rotation = this_transfo;
		} else {
			//// first rotation ////
			// prec transfo
			// there isn't any prec transfo
			// next transfo
			if (_first_translation != -1)
				_transfos[_first_translation].prec = this_transfo;			
			// prec link
			_transfos[this_transfo].prec = -1;
			// next link
			_transfos[this_transfo].next = _first_translation;
			// first
			_first_rotation = this_transfo;
			// last
			_last_rotation = this_transfo;
			// the first pointer must move to this transfo
			_first_transfo = this_transfo;
		}
	}
	if (debug > 1) {
		if (transl_or_not_rotation)
			g_print("  nb transfos = %d, translation added at %d\n ",
			_nb_transfo,this_transfo);
		else 
			g_print("  nb transfos = %d, rotation added at %d\n ",
			_nb_transfo,this_transfo);
		gint i = _first_transfo;
		if (_nb_transfo) while (i != -1) {
			g_print("  ");
			if (i == this_transfo) g_print("+[");
			g_print("%d<%d:",_transfos[i].prec,i);
			if (_transfos[i].fx >= trans_U && _transfos[i].fx <= trans_LU)
				g_print("T");
			else if (_transfos[i].fx >= trans_XP && _transfos[i].fx <= trans_ZN)
				g_print("R");
			else if (_transfos[i].fx == trans_FALL)
				g_print("F");
			else g_print("?");
			g_print(">%d",_transfos[i].next);
			if (i == this_transfo) g_print("]");
			i = _transfos[i].next;
		}
		if (_nb_transfo) g_print("\n");
		g_print("  1stX=%d, 1stR=%d, lR=%d, 1stT=%d, lT=%d\n",
			 _first_transfo, _first_rotation, _last_rotation,
			 _first_translation, _last_translation);
	}
	return TRUE;
}

void CGame::clear_transfos(void)
{
	_nb_transfo = 0;
	_first_translation = -1;
	_last_translation = -1;
	_first_rotation = -1;
	_last_rotation = -1;
	_first_transfo = -1;
	for (gint i=0; i<NB_TRANSFO_MAX; i++) {
		_transfos[i].fx = trans_NONE;
		_transfos[i].t = 0;
		_transfos[i].next = -1;
		_transfos[i].prec = -1;
	}	
}

void CGame::compute_transfos(gfloat dt)
{
	if (debug > 2) g_print("CGame::compute_transfos %.2f\n",dt);
	if (_nb_transfo <= 0) return; // nothing to do
	gint index = _first_transfo;
	gint next = -1;
	gint i;
	gfloat gap = dt * HALF_OF_PI;
	gfloat cosap = cos(gap);
	gfloat sinap = sin(gap);
	gfloat t;
	_xing_piece = _roted_piece;
	while (index != -1) { 
		if (_transfos[index].fx == trans_FALL)
			_transfos[index].t += (dt*(_level+1)*_falling_speed)/20.0;
		else {
			if (_settings->_speed == CSettings::speed_INSTANT)
				_transfos[index].t = 1;
			else if (_settings->_speed == CSettings::speed_FAST)
				_transfos[index].t += dt*(_level+1)*1.5;
			else // speed = NORMAL
				_transfos[index].t += dt*(_level+1)/1.5;
		}
		t = _transfos[index].t;
		if (debug > 2) g_print("  compute transfo %d at %d: t=%.2f\n",_transfos[index].fx,index,t);
		if (t > 1.0) t = 1.0; // completion of the transfo
		// transfo
		switch (_transfos[index].fx) {
			case trans_FALL: { // up
				_xing_piece.translate_z(t, &_roted_piece);
				break;
			}
			case trans_U: { // up
				_xing_piece.translate_y(-t, &_roted_piece);
				break;
			}
			case trans_R: { // right
				_xing_piece.translate_x(t, &_roted_piece);
				break;
			}
			case trans_D: { // down
				_xing_piece.translate_y(t, &_roted_piece);
				break;
			}
			case trans_L: { // left
				_xing_piece.translate_x(-t, &_roted_piece);
				break;
			}
			case trans_XP: { // x +
				_xing_piece.rotate_x(t, &_roted_piece);
				break;
			}
			case trans_XN: { // x -
				_xing_piece.rotate_x(-t, &_roted_piece);
				break;
			}
			case trans_YP: { // y +
				_xing_piece.rotate_y(t, &_roted_piece);
				break;
			}
			case trans_YN: { // y -
				_xing_piece.rotate_y(-t, &_roted_piece);
				break;
			}
			case trans_ZP: { // z +
				_xing_piece.rotate_z(t, &_roted_piece);
				break;
			}
			case trans_ZN: { // z -
				_xing_piece.rotate_z(-t, &_roted_piece);
				break;
			}
			default: if (debug)	g_print("** ERROR **: CGame::invalid transfo at %d\n", index);
		} // switch transfo
		next = _transfos[index].next; // save next now because
			// it may be cleared in the next line:
		if (t == 1.0) remove_transfo(index);
		index = next;
	} // loop transfo queue
	// apply translations
	_xing_piece.apply_translations();
}

/*static*/ void CGame::create_pieces(void)
{
	if (debug) g_print("CGame::create_pieces\n");
	g_assert(nb_pieces < NB_PIECES_MAX);
	gint i;
	gint d = 0;
	gint ie = 0;
	gint im = 0;
	gint ih = 0;
	for (i=0; i< NB_PIECES_MAX; i++) {
		if (piece_descriptors[i] == NULL
		 || !g_strcasecmp(piece_descriptors[i], "EOL")) break;
		pieces[nb_pieces] = g_new(CPiece,1);
		if (d=pieces[nb_pieces]->create(piece_descriptors[i])) {
			pieces_set[CSettings::diff_HARD][ih++] = nb_pieces;
			if (d <= 2) pieces_set[CSettings::diff_MEDIUM][im++] = nb_pieces;
			if (d == 1) pieces_set[CSettings::diff_EASY][ie++] = nb_pieces;
			nb_pieces++;
	}	}
	nb_pieces_by_diff[CSettings::diff_EASY] = ie;
	nb_pieces_by_diff[CSettings::diff_MEDIUM] = im;
	nb_pieces_by_diff[CSettings::diff_HARD] = ih;
	if (debug) g_print("CGame::create_pieces, %d pieces created\n",nb_pieces);
	if (debug) g_print("  easy set=%d, medium=%d, hard=%d\n",
		nb_pieces_by_diff[CSettings::diff_EASY],
		nb_pieces_by_diff[CSettings::diff_MEDIUM],
		nb_pieces_by_diff[CSettings::diff_HARD]);
}

/*static*/ void CGame::destroy_pieces(void)
{
	if (debug) g_print("CGame::destroy_pieces\n");
	for (gint i=0; i<CGAME_NB_PIECES_MAX; i++)
		if (pieces[i] != NULL) {
		    pieces[i]->free_memory(); // remove cubes, arretes & pointes arrays
			g_free(pieces[i]);
			pieces[i] = NULL;
}	}

gboolean CGame::does_piece_fit_in_the_world(CPiece *p)
{
	if (debug > 2) g_print("CGame::does_piece_fit_in_the_world\n");
	if (_settings == NULL) return FALSE;
	if (p->_dx > _settings->_w || p->_dx > _settings->_h) return FALSE;
	if (p->_dy > _settings->_w || p->_dy > _settings->_h) return FALSE;
	if (p->_dz > _settings->_w || p->_dz > _settings->_h) return FALSE;
	return TRUE;
}

/* fall
 * RET: TRUE if the world needs to be re-drawn
*/
gboolean CGame::fall(gint speed)
{
	if (debug > 2) g_print("CGame::fall, falling = %d, speed = %d\n",
		_falling, _falling_speed);
	if (!_falling) {
		if (add_transfo(trans_FALL)) {
			// validate move
			_falling = TRUE;
			return FALSE;
		} else {
			gint dscore = 0;
			// the fall isn't valid
			// stick piece and get another one
			set_piece();
			gint completed_layers = 0;
			if (get_next_piece()) completed_layers = remove_completed_layers();
			// compute new level
			gint old_level = _level;
			_level = (gint)(div(_layers,INCREMENT_LEVEL_EVERY_N_LAYERS).quot+1);
			if (_level > old_level)
				// sound
				if (_settings->_sound_support)
					csound_play_for_event(CSound::snd_LEVEL);
			// compute score due to layers done
			if (completed_layers) {
				dscore = 20								 // +20
					+ completed_layers*completed_layers  // +1,4,9,16,25
					* _settings->_difficulty			 // *1,2,3
					* 30								 // *30
					/ (_settings->_d - MIN_WORLD_DEPTH + 1); // /1-16
					
				if (completed_layers > 3) {
					// tetris bonus !
					dscore += _settings->_difficulty*500*(completed_layers-3);
						// layers=   4      5
						// diff=1: + 500  +1000
						//     =2: +1000  +2000
						//     =3: +1500  +3000
					// sound
					if (_settings->_sound_support)
						csound_play_for_event(CSound::snd_SCORE);
				}
				_score += dscore;
				if (debug) g_print("  layer bonus adds %d to score\n",dscore);
			}
			if (debug) g_print("CGame::fall exits, level=%d, score=%d\n",
				_level,_score);
			return TRUE;
	}	}
	return FALSE;
}

void CGame::fast_fall(void) {
	_falling_speed = CGAME_FALLING_SPEED; 
	// sound
	if (_settings->_sound_support)
		csound_play_for_event(CSound::snd_FALL);
};

void CGame::free_memory(void)
{
	if (debug > 2) g_print("CGame::free_memory\n");
	if (_name != NULL) {
		g_free(_name);
		_name = NULL;
	}
	if (_date != NULL) {
		g_free(_date);
		_date = NULL;
	}
	if (_settings != NULL) {
		g_free(_settings);
		_settings = NULL;
	}
}

void CGame::game_lost(void)
{
	if (debug) g_print("CGame::game_lost\n");
	set_piece();
	_state = CGame::state_LOST;
	// sound
//	if (_settings->_sound_support)
//		gnome_sound_play(_settings->_sound_over);
}

/* get_next_piece
 * RET: FALSE = game lost, TRUE otherwise
*/
gboolean CGame::get_next_piece(void)
{
	if (debug) g_print("CGame::get_next_piece, diff = %s\n",
		CSettings::diff_descriptors[_settings->_difficulty]);

	// get next piece
	if (_next_piece == -1) _piece = *(pieces[get_random_piece()]);
	else _piece = *(pieces[_next_piece]);
	_next_piece = get_random_piece();

	// add translations to have the cubes even with the cubes of the world
	_piece._tx = _hw;
	if (!div(_settings->_w,2).rem) _piece._tx -= 0.5;
	_piece._ty = _hh;
	if (!div(_settings->_h,2).rem) _piece._ty -= 0.5;
	_piece._tz = 0.5; //0.5 + _piece._dz -1 ;
	if (debug > 1) g_print("  hw=%.2f, hh=%.2f, added translations %.2f,%.2f,%.2f\n",
		_hw, _hh, _piece._tx, _piece._ty, _piece._tz);
	if (debug > 1) cout << "  next piece = " << _next_piece  << endl;
	
	// init pieces
	_roted_piece = _piece;
	_xing_piece = _piece;
	_xing_piece.apply_translations(); // apply translations to the moving piece
	_xed_piece = _piece;

	// clear transfo queue
	_falling = FALSE;
	_falling_speed = 1;
	clear_transfos();
	
	// test game status
	if (!is_valid_location(&_piece)) {
		game_lost();
		return FALSE;
	}
	return TRUE;
}

gint CGame::get_random_piece(void)
{
	if (debug > 1) g_print("CGame::get_random_piece\n");
	gint r;
	gint p;
	gint range = nb_pieces_by_diff[_settings->_difficulty];
//	time_t t;
//  ::time(&t);
//  srand((guint)t);
	
	r = (gint)((gfloat)(range)*rand()/(RAND_MAX+1.0));
	p = pieces_set[_settings->_difficulty][r];
	gint safe_counter = 40;
	while (!does_piece_fit_in_the_world(pieces[p])) {
		safe_counter--;
		if (safe_counter == 0) {
			// the while loop is taking too long, let's take the first piece
			// available
			for (gint i=0; i<range; i++) {
				if (i+1 == range) {
					// there's no pieces available, we should never reach
					// this piece of code. bug !
					if (debug) g_print("** ERROR **: CGame::get_random_piece, can't find a piece that fits in the world\n");
					p = 0;
					break;
				}
				r = i;
				p = pieces_set[_settings->_difficulty][r];
				if (does_piece_fit_in_the_world(pieces[p])) break;
			}
		} else { // safe counter > 0
			// try once more
			r = (gint)((gfloat)(range)*rand()/(RAND_MAX+1.0));
			p = pieces_set[_settings->_difficulty][r];
		}
	} // while
	g_assert(pieces[p] != NULL);
	return p;
}
	
void CGame::init(gboolean keep_current_settings)
{
	if (debug) g_print("CGame::init\n");

	if (!keep_current_settings || _settings == NULL) {
		// game id
		set_game_id();

		// name
		set_name();
	
		// load settings
		if (_settings == NULL) _settings = g_new0(CSettings,1);
		*_settings = *settings;
		if (debug) cout << *_settings;
		_hw = _settings->_w/2.0;
		_hh = _settings->_h/2.0;
	}
	// init world
	gint x,y,z;
	for (x=0; x<MAX_WORLD_WIDTH+2; x++)
		for (y=0; y<MAX_WORLD_HEIGHT+2; y++)
			for (z=0; z<MAX_WORLD_DEPTH+4; z++)
				_world[x][y][z] = TRUE;
	for (x=1; x<_settings->_w+1; x++)
		for (y=1; y<_settings->_h+1; y++)
			for (z=1; z<MAX_WORLD_DEPTH+4; z++)
				_world[x][y][z] = FALSE;	

	// init layer counts
	for (z=1; z<MAX_WORLD_DEPTH+4; z++)
		_layer[z] = 0;

	_score = 0;
	_layers = 0;
	_cubes = 0;
	_blocs = 0;
	_highest_cube = 0;
	reset_level();
	_next_piece = -1;
	if (!keep_current_settings || _settings == NULL) get_next_piece();
}

gboolean CGame::is_valid_location(CPiece *piece)
{
	if (debug > 1) g_print("CGame::is_valid_location\n");
	gboolean valid = TRUE;
	gint cx;
	gint cy;
	gint cz;
	for (gint i=0; i<piece->_nb_cubes; i++) {
		cx = (gint)(piece->_cubes[i]->x + piece->_tx + 1);
		cy = (gint)(piece->_cubes[i]->y + piece->_ty + 1);
		cz = (gint)(_settings->_d - piece->_cubes[i]->z - piece->_tz + 1);
		if (debug > 2) g_print(", %d=(%d,%d,%d)",i,cx,cy,cz);	
		if (_world[cx][cy][cz]) {
			if (debug > 2) g_print(" <- it's occupied here");
			valid = FALSE; 
			break;
		}
	}
	if (debug > 2) g_print("\n");
	// the move is valid
	if (debug > 1) g_print("  is_valid_location exits with %d\n",valid);
	return valid;	
}

gboolean CGame::is_valid_transfo(gint transfo)
{
	if (debug > 1) g_print("CGame::is_valid_transfo (%d)\n",transfo);
	CPiece tmpp;
	tmpp = _xed_piece;
	switch (transfo) {
		case trans_FALL: tmpp.translate_z(1.0, NULL, FALSE, FALSE);
			break;
		case trans_U: tmpp.translate_y(-1.0, NULL, FALSE, FALSE);
			break;
		case trans_R: tmpp.translate_x(1.0, NULL, FALSE, FALSE);
			break;
		case trans_D: tmpp.translate_y(1.0, NULL, FALSE, FALSE);
			break;
		case trans_L: tmpp.translate_x(-1.0, NULL, FALSE, FALSE);
			break;
		case trans_XP: tmpp.rotate_x(1.0, NULL, FALSE, TRUE);
			break;
		case trans_XN: tmpp.rotate_x(-1.0, NULL, FALSE, TRUE);
			break;
		case trans_YP: tmpp.rotate_y(1.0, NULL, FALSE, TRUE);
			break;
		case trans_YN: tmpp.rotate_y(-1.0, NULL, FALSE, TRUE);
			break;
		case trans_ZP: tmpp.rotate_z(1.0, NULL, FALSE, TRUE);
			break;
		case trans_ZN: tmpp.rotate_z(-1.0, NULL, FALSE, TRUE);
			break;
		default: if (debug)	g_print("** ERROR **: CGame::is_valid_transfo %d\n",
			index);
	}
	if (debug > 2)
		g_print("  1/2w=%.2f, 1/2h=%.2f, tx=%.2f, ty=%.2f, tz=%.2f,\
ncubes=%d\n", _hw, _hh, tmpp._tx, tmpp._ty, tmpp._tz, tmpp._nb_cubes);
	gboolean valid = is_valid_location(&tmpp);
	// the move is valid
	if (valid) _xed_piece = tmpp;
	else {
		// try to do the move even if invalid
		// for example, translate the piece away from the wall making the
		// move invalid
		if (debug > 1) g_print("  searching a valid adjusted translation\n");
		CPiece tmpp2;
		switch (transfo) {
			case trans_XP: case trans_XN: {
				// one translation
				tmpp2 = tmpp;
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_D, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_U, FALSE/*don't check validity*/);
					break;
				}
				// 2 translations
				tmpp2 = tmpp;
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_D, FALSE/*don't check validity*/);
					add_transfo(trans_D, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_U, FALSE/*don't check validity*/);
					add_transfo(trans_U, FALSE/*don't check validity*/);
					break;
				}
				// no valid move
				break;
			}
			case trans_YP: case trans_YN: {
				// one translation
				tmpp2 = tmpp;
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_R, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_L, FALSE/*don't check validity*/);
					break;
				}
				// 2 translations
				tmpp2 = tmpp;
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_R, FALSE/*don't check validity*/);
					add_transfo(trans_R, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_L, FALSE/*don't check validity*/);
					add_transfo(trans_L, FALSE/*don't check validity*/);
					break;
				}
				// no valid move
				break;
			}
			case trans_ZP: case trans_ZN: {
				// 1 translation
				tmpp2 = tmpp;
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_U, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_R, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_D, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_L, FALSE/*don't check validity*/);
					break;
				}
				// 2 translations
				tmpp2 = tmpp;
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				tmpp2.translate_y(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_U, FALSE/*don't check validity*/);
					add_transfo(trans_U, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				tmpp2.translate_x(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_R, FALSE/*don't check validity*/);
					add_transfo(trans_R, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				tmpp2.translate_y(1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_D, FALSE/*don't check validity*/);
					add_transfo(trans_D, FALSE/*don't check validity*/);
					break;
				}
				tmpp2 = tmpp;
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				tmpp2.translate_x(-1.0, NULL, FALSE, FALSE);
				if (is_valid_location(&tmpp2)) {
					valid = TRUE;
					_xed_piece = tmpp2;
					add_transfo(trans_L, FALSE/*don't check validity*/);
					add_transfo(trans_L, FALSE/*don't check validity*/);
					break;
				}
				// no valid move
				break;
			} // case
		} // switch trans_X
	} // if valid
	if (debug > 1) g_print("  is_valid_transfo exits with %d\n",valid);
	return valid;
}

/* remove_completed_layers
 * RET: number of completed layers
*/
gint CGame::remove_completed_layers(void)
{
	if (debug > 1) g_print("CGame::remove_completed_layers\n");
	gint complete_count = _settings->_w * _settings->_h;
	gint completed_layer = 0;
	for (gint z=_settings->_d+1; z>0; z--) {
		if (_layer[z] >= complete_count) {
			// sound
			if (_settings->_sound_support)
				csound_play_for_event(CSound::snd_LAYER);
			completed_layer++;
			// remove layer
			for (gint zz=z; zz<_settings->_d+1; zz++) {
				// move above layers down
				for (gint x=1; x<_settings->_w+2; x++)
					for (gint y=1; y<_settings->_h+2; y++)
						_world[x][y][zz] = _world[x][y][zz+1];
				_layer[zz] = _layer[zz+1];
	}	}	}
	_layers += completed_layer;
	_highest_cube -= completed_layer;
	if (debug) g_print("CGame::remove_completed_layers: %d layers completed, \
%d layers total\n", completed_layer, _layers);
	return completed_layer;	
}

void CGame::remove_transfo(gint index)
{
	if (debug > 1) g_print("CGame::remove_transfo (%d)\n",index);
	if (index < 0 || index > NB_TRANSFO_MAX) return;

	// prec transfo
	if (_transfos[index].prec != -1)
		_transfos[_transfos[index].prec].next = _transfos[index].next;
	// next transo
	if (_transfos[index].next != -1)
		_transfos[_transfos[index].next].prec = _transfos[index].prec;
	// first transfo
	if (index == _first_transfo)
		_first_transfo = _transfos[index].next;
	// no more transfo of this kind
	if (index == _first_rotation && index == _last_rotation) {
		_first_rotation = -1;
		_last_rotation = -1;
	} else if (index == _first_translation && index == _last_translation) {
		_first_translation = -1;
		_last_translation = -1;
	} else {
		// first
		if (index == _first_rotation)
			_first_rotation = _transfos[index].next;
		else if (index == _first_translation)
			_first_translation = _transfos[index].next;
		// last
		if (index == _last_rotation)
			_last_rotation = _transfos[index].prec;
		else if (index == _last_translation)
			_last_translation = _transfos[index].prec;
	}
	// data & links
	if (_transfos[index].fx == trans_FALL) _falling = FALSE;
	_transfos[index].fx = trans_NONE;
	_transfos[index].next = -1;
	_transfos[index].prec = -1;
	_nb_transfo--;
	if (_nb_transfo < 0) _nb_transfo = 0;

	if (debug > 1) {
		gint this_transfo = index;
		g_print("  nb transfos = %d, transfo removed from %d\n ",
			_nb_transfo,this_transfo);
		gint i = _first_transfo;
		if (_nb_transfo) while (i != -1) {
			g_print("  ");
			if (i == this_transfo) g_print("+[");
			g_print("%d<%d:",_transfos[i].prec,i);
			if (_transfos[i].fx >= trans_U && _transfos[i].fx <= trans_LU)
				g_print("T");
			else if (_transfos[i].fx >= trans_XP && _transfos[i].fx <= trans_ZN)
				g_print("R");
			else if (_transfos[i].fx == trans_FALL)
				g_print("F");
			else g_print("?");
			g_print(">%d",_transfos[i].next);
			if (i == this_transfo) g_print("]");
			i = _transfos[i].next;
		}
		if (_nb_transfo) g_print("\n");
		g_print("  1stX=%d, 1stR=%d, lR=%d, 1stT=%d, lT=%d\n",
			 _first_transfo, _first_rotation, _last_rotation,
			 _first_translation, _last_translation);
	}
}

/* set_game_id
*/
gint32 CGame::set_game_id(void)
{
	if (debug) g_print("CGame::set_game_id\n");
	// 4 bytes = gint32 = 3 bytes:time; 1 byte:random
	time_t t;
	::time(&t);
	gint32 r3 = (gint32)t; 
	r3 = (0x00FFFFFF & r3) * 0x100; // shift 1 byte <-
	gint32 r1 = (gint32)((gfloat)(0xFF)*rand()/(RAND_MAX+1.0)) & 0x000000FF;
	r3 = r3 + r1;
	_id = r3;
	return r3;
}

gchar *CGame::set_name(gchar *str)
{
	if (_name != NULL) {
		g_free(_name);
		_name = NULL;
	}
	if (str == NULL) _name = g_strndup(g_get_real_name(),MAX_NAME_LEN);
	else _name = g_strndup(str,MAX_NAME_LEN);
	return _name;
}

void CGame::set_piece(void)
{
	if (debug) g_print("CGame::set_piece\n");
	gint cx;
	gint cy;
	gint cz;
	gint nb_cubes = _xed_piece._nb_cubes;
	gint _old_highest_cube = _highest_cube;
	gboolean made_hole = FALSE;
	gint underneath_cubes[nb_cubes][3];
	for (gint i=0; i<nb_cubes; i++) {
		cx = (gint)(_xed_piece._cubes[i]->x + _xed_piece._tx + 1);
		cy = (gint)(_xed_piece._cubes[i]->y + _xed_piece._ty + 1);
		cz = (gint)(_settings->_d-_xed_piece._cubes[i]->z-_xed_piece._tz+1);
		_world[cx][cy][cz] = TRUE;
		underneath_cubes[i][0] = cx;
		underneath_cubes[i][1] = cy;
		underneath_cubes[i][2] = cz-1;
		if (cz > _settings->_d) {
			_world[cx][cy][cz] = FALSE; // avoids pbs with
			// long pieces going higher than the top of the world
			cz = _settings->_d;
		}
		_layer[cz]++;
		if (cz > _highest_cube) _highest_cube = cz;
	}
	// check for hole underneath the piece
	for (gint i=0; i<nb_cubes; i++)
		if (underneath_cubes[i][2] < _settings->_d) // not on top or above
			if (!_world[underneath_cubes[i][0]]
					  [underneath_cubes[i][1]]
					  [underneath_cubes[i][2]]) {
				made_hole = TRUE;
				continue; //  exit for()
	}
	// sound
	if (_settings->_sound_support)
		if (made_hole) csound_play_for_event(
			CSound::snd_SCREWD,	CSound::snd_LAND);
		else csound_play_for_event(CSound::snd_LAND);
	if (_settings->_sound_support
	 && _highest_cube > _old_highest_cube // must be higher
	 && _highest_cube > _settings->_d-4 // too close to the top
	 && _highest_cube < _settings->_d-1) // but not at the top (game ~lost)
		csound_play_for_event(CSound::snd_DANGER);
	// stats
	_blocs++;
	_cubes += nb_cubes;
	// compute delta score
	gint dscore =
	    5 * ( _xed_piece._class		// 5*( 1,2,3
			+ nb_cubes				//    +1..5
			+ _level				//    +1..oo
			+ 5)					//    +5)
		* _settings->_difficulty	// * 1,2,3
		/ _settings->_d;			// / 5..20
	if (debug) g_print("  set piece adds %d to score\n",dscore);
	_score += dscore;
}

void CGame::start(void)
{
	if (debug) g_print("CGame::start\n");
	_state = state_PLAYING;
	reset_timer(); // clock
}

gchar *CGame::stringify_timer(void)
{
//	if (debug > 2) g_print("CGame::stringify time\n");
	if (_timer > (999*60+59)*60+59) return g_strdup("888:88:88");
	gint seconds = div((int)_timer,60).rem;
	int minutes = div((int)_timer,60).quot;
	gint heures = div(minutes,60).quot;
	minutes = div(minutes,60).rem;
	return g_strdup_printf("%03u:%02u:%02u",heures,minutes,seconds);
}
		
// IO STUFF ///////////////////////////////////////////////////////////////////

/* DO NOT REMOVE - IT'LL BE USEFUL LATER

void CGame::dialog_error_reading(ifstream &s, const gchar *name, gchar *m) {
	if (debug) g_print("CGame::dialog error reading\n");
	if (s.is_open()) s.close();
	gchar *msg = g_strdup_printf(_(
		"An error occured while reading data from file:\n \
		%s:\n \
		%s.\n \
		Operation aborted"),
		name, m);
	GtkWidget *dialog = gnome_error_dialog(msg);
	gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
	g_free(msg);
}

void CGame::dialog_error_writing(ofstream &s, const gchar *name, gchar *m) {
	if (debug) g_print("CGame::dialog error wrinting\n");
	if (s.is_open()) s.close();
	gchar *msg = g_strdup_printf(_(
		"An error occured while writing data to file:\n \
		%s:\n \
		%s.\n \
		Operation aborted"),
		name, m);
	GtkWidget *dialog = gnome_error_dialog(msg);
	gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
	g_free(msg);
}*/

/* read_game
 * RET: 0 -> ok
 *      1 -> empty file / EOF
 *		3 -> safe counter failure
 *		4 -> read invalid value
 *		5 -> read unknown key
 *      6 -> error opening
 *		- -> other
*/
/*
gint CGame::read_game(const gchar *filename,
	gboolean pos_beginning_or_not_current)
{
	if (pos_beginning_or_not_current) CGame::_pos = 0L;
	long pos = CGame::_pos;
	if (debug) g_print("CGame::read_game (>%s<,%d) @ pos = %d\n",\
		filename,pos_beginning_or_not_current,pos);
	gint ret = read_at(&pos, filename);
	_pos = pos;
	if (debug) g_print("CGame::read_game: exit elvl %d\n",ret);	
	return ret;
}

gint CGame::read_at(long *pos, const gchar *filename)
{
	if (debug > 2) g_print("CGame::read_at >%s< %d\n",filename,*pos);

	// open file
	ifstream fin(filename);
	if (!fin.is_open()) {
		dialog_error_reading(fin, filename, _("File opening failed"));
		return 6;
	}
	const gint MAX_BUF = 1024;
	gchar buf[MAX_BUF] = "";		
	gchar buf2[MAX_BUF];
	const gint safe_counter_init = 12;
	gint safe_counter = safe_counter_init;
	fin.seekg(*pos);

	// check for EOF
	fin >> buf;
	if (fin.eof()) {
		if (debug) g_print("CGame::read_at: EOF\n");
		return 1;
	}
	else fin.seekg(*pos); // rewind

	// reset settings
	free_memory();	
	if (_settings == NULL) _settings = g_new0(CSettings,1);
	
	// read record
	fin.seekg(*pos); // rewind
	safe_counter = safe_counter_init;
	fin >> buf;
	while (fin.good()) {
		if (--safe_counter < 0) return 3;
		if (debug > 2) g_print("  buffer read = >%s<\n",buf);
		if (!g_strcasecmp(str_id, buf)) {
			fin >> _id;
			if (debug > 2) g_print("  id found: read = %d\n",_id);		
		} else if (!g_strcasecmp(str_name, buf)) {
			fin.getline(buf2, MAX_BUF);
			gint offset = 0;
			while (buf2[offset] == ' ') { offset++; };
			if (buf2[offset] == '\0') offset = 0;
			_name = g_strndup(&buf2[offset], MAX_NAME_LEN);
			if (debug > 2) g_print("  name found: read = >%s<\n",_name);		
		} else if (!g_strcasecmp(str_timer, buf)) {
			fin >> _timer;
			if (debug > 2) g_print("  time found: read = %d\n",_timer);		
		} else if (!g_strcasecmp(str_date, buf)) {
			fin.getline(buf2, MAX_BUF);
			gint offset = 0;
			while (buf2[offset] == ' ') { offset++; };
			if (buf2[offset] == '\0') offset = 0;
			_date = g_strndup(&buf2[offset], MAX_HOF_S_LEN);
			if (debug > 2) g_print("  date found: read = >%s<\n",_date);		
		} else if (!g_strcasecmp(str_level, buf)) {
			fin >> _level;
			if (debug > 2) g_print("  level found: read = %d\n",_level);		
		} else if (!g_strcasecmp(str_score, buf)) {
			fin >> _score;
			if (debug > 2) g_print("  score found: read = %d\n",_score);		
		} else if (!g_strcasecmp(str_layers, buf)) {
			fin >> _layers;
			if (debug > 2) g_print("  layers found: read = %d\n",_layers);		
		} else if (!g_strcasecmp(CSettings::str_diff, buf)) {
			fin >> _settings->_difficulty;
			if (debug > 0) g_print("  settings difficulty found: read = %d\n",
				_settings->_difficulty);		
		} else if (!g_strcasecmp(CSettings::str_width, buf)) {
			fin >> _settings->_w;
			if (debug > 2) g_print("  settings width found: read = %d\n",
				_settings->_w);		
		} else if (!g_strcasecmp(CSettings::str_height, buf)) {
			fin >> _settings->_h;
			if (debug > 2) g_print("  settings height found: read = %d\n",
				_settings->_h);		
		} else if (!g_strcasecmp(CSettings::str_depth, buf)) {
			fin >> _settings->_d;
			if (debug > 2) g_print("  settings depth found: read = %d\n",
				_settings->_d);		
		} else if (!g_strncasecmp(str_eog, buf, strlen(str_eog))) {
			if (debug > 2) g_print("  EOG found\n");		
			break; // exit while loop
		} else {
			if (debug) g_print("Read an invalid key : >%s<",buf);
			gchar *msg = g_strdup_printf(_("The file containing the Hall Of Fame is corrupted"));
			dialog_error_reading(fin, filename, msg);
			g_free(msg);
			return 5;
		}
		fin >> buf;
	}
	*pos = fin.tellg();
	return 0;
}

gint CGame::write_game(const gchar* filename, gboolean trunc_or_not_app)
{
	if (debug) g_print("CGame::write_game >%s< , %d\n",
		filename,trunc_or_not_app);

	ofstream fout;
	if (trunc_or_not_app) {
		fout.open(filename, ios::out | ios::trunc);
	} else {
		fout.open(filename, ios::out | ios::app);
	}
	g_assert(fout.is_open());
	// data
	fout << str_id << " " << _id << endl;
	if (!fout.good()) {
		if (debug) {
			g_print("CGame:write_game: something wrong:");
			if ((fout).eof()) g_print(" EOF");
			if ((fout).fail()) g_print(" FAIL");
			if ((fout).bad()) g_print(" BAD");
			g_print("; exit elvl 3\n");
		}
		dialog_error_writing(fout, filename, _("Writing operation failed"));
		return 3;			
	}
	if (_settings != NULL) {
		fout << CSettings::str_diff; 
		fout << " " << _settings->_difficulty << endl;
		fout << CSettings::str_depth  << " " << _settings->_d << endl;
		fout <<	CSettings::str_width  << " " << _settings->_w << endl;
		fout << CSettings::str_height << " " << _settings->_h << endl;
	}
	fout <<	str_score  << " " << _score  << endl;
	fout <<	str_level  << " " << _level  << endl;
	fout <<	str_layers << " " << _layers << endl;	
	fout << str_timer  << " " << _timer  << endl;
	if (_date != NULL) fout << str_date << " " << _date << endl;
	if (_name != NULL) fout << str_name << " " << _name << endl;
	// end of record
	fout << str_eog << endl;
	if (fout.is_open()) fout.close();
	if (debug) g_print("CGame::write_game: exit elvl 0\n");
	return 0;
}
*/
