/* gno3dtet: 3D Tetris game for GNOME
 * Copyright (C) 1999, 2000, 2001 Sebastien Nicoud <seb@eseb.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/

/* create_pieces must be called after the first instance of CGame is created
 * destroy_pieces() must be called after the last instance of CGame is destroyed
*/

#ifndef _CGAME
#define _CGAME

#include "common.h"
#include "cpiece.h"

#define CGAME_NB_PIECES_MAX			40
#define CGAME_NB_TRANSFO_MAX		20
#define CGAME_FALLING_SPEED_SLOW	1
#define CGAME_FALLING_SPEED			2000

class CGame
{
// friends
friend class CHof;

// typedefs, consts
private:
	static const gint 	NB_PIECES_MAX;
	static const gint 	NB_TRANSFO_MAX;
	static gint  		nb_pieces;
	static const gchar *piece_descriptors[CGAME_NB_PIECES_MAX];
	static gint 		pieces_set[4][CGAME_NB_PIECES_MAX];
	static gint 		nb_pieces_by_diff[4];
public:
	static CPiece 	   *pieces[CGAME_NB_PIECES_MAX];

public:
	enum {
		state_NEW,			// the game isn't started yet
		state_DEMO,			// demo mode // not implemented
		state_PLAYING,		// playing
		state_PAUSED,		// game paused
		state_LOST,			// game lost
		state_DONE			// game is over
	};
	enum {
		trans_NONE,
		trans_FALL,
		trans_U,
		trans_RU,
		trans_R,
		trans_DR,
		trans_D,
		trans_DL,
		trans_L,
		trans_LU,
		trans_XP,
		trans_XN,
		trans_YP,
		trans_YN,
		trans_ZP,
		trans_ZN,
		trans_LAST // boundary marker
	};
	typedef struct _t_transfo {
		gint8	fx;
		gdouble	t;
		gint8	next;
		gint8	prec;
	} t_transfo;
//	static const gchar *str_id; // DON'T REMOVE
//	static const gchar *str_level;
//	static const gchar *str_score;
//	static const gchar *str_layers;
//	static const gchar *str_timer;
//	static const gchar *str_name;
//	static const gchar *str_date;
//	static const gchar *str_eog;

// attributes
private:
	gint32		 _id; // game id
	gchar		*_name; // player name
	gchar		*_date; // current date
	gint32		 _timer; // clock
	
	gint8		 _nb_transfo;
	gint8		 _first_translation;
	gint8		 _last_translation;
	gint8		 _first_rotation;
	gint8		 _last_rotation;
	gint8		 _first_transfo;
	t_transfo	 _transfos[CGAME_NB_TRANSFO_MAX];
	
//	long		 _pos; // for I/O - DON't REMOVE
	gint		 _dscore;
	
public:
	gint		 _level; // game level
	gint32		 _score; // score
	gint		 _layers; // nb of completed layers
	gint		 _cubes; // nb of cubes
	gint		 _blocs; // nb of blocs
	gint		 _state;
	CSettings	*_settings;
	gfloat		 _hw;
	gfloat		 _hh;
	gboolean	 _world[MAX_WORLD_WIDTH+2]	// TRUE = full bloc
					   [MAX_WORLD_HEIGHT+2]	// FALSE = empty bloc
					   [MAX_WORLD_DEPTH+4];
	gint		 _layer[MAX_WORLD_DEPTH];
	gboolean	 _smooth;
	gboolean	 _falling;
	gint		 _falling_speed;
	CPiece		 _piece; // index of current piece, in pieces[] array
	gint		 _next_piece; // index of next piece, in pieces[] array
	CPiece		 _roted_piece; // fully rotated piece, before current transfos
	CPiece		 _xing_piece; // transforming piece
	CPiece		 _xed_piece; // fully transformd piece, current transfos includd
	gint		 _highest_cube;
		
// contructor
public:
	CGame(); // constructeur par defaut
	CGame(const CGame &ind); // constructeur de copy
	
// destructor
	~CGame();

// operators
public:
	CGame &operator= (const CGame &ind); // affectation de recopie
	friend ostream &operator<< (ostream &os, CGame &ind); // output stream

// methods
public:
	gboolean add_transfo(gint transfo, gboolean check_validity = TRUE);
	void compute_transfos(gfloat dt);
	static void create_pieces(void);
	static void destroy_pieces(void);
	gboolean fall(gint speed = 1);
	void fast_fall(void);
	void inc_level(void) { _level++; };
	void inc_timer(void) { _timer++; };
	void init(gboolean keep_current_settings = FALSE);
	gint remove_completed_layers(void);
	void remove_transfo(gint index);
	void reset_level(void) { _level = 1; };
	void reset_timer(void) { _timer = 0; };
	gchar *set_name(gchar *str = NULL);
	void start(void);
	gchar *stringify_timer(void);
	
private:
	void clear_transfos(void);
	gboolean does_piece_fit_in_the_world(CPiece *p);
	void free_memory(void);
	void game_lost(void);
	gboolean get_next_piece();
	gint get_random_piece(void);
	gboolean is_valid_location(CPiece *piece);
	gboolean is_valid_transfo(gint transfo);
	gint32 set_game_id(void);
	void set_piece(void);

// i/o methods // DON'T REMOVE
//public:
//	gint read_game(const gchar* filename,
//			gboolean pos_beginning_or_not_current = TRUE);
//	gint write_game(const gchar* filename, gboolean trunc_or_not_app = TRUE);
//private:
//	void dialog_error_reading(ifstream &s, const gchar *name, gchar *m);
//	void dialog_error_writing(ofstream &s, const gchar *name, gchar *m);
//	gint read_at(long *pos, const gchar* filename);
};

#endif
