/* gno3dtet: 3D Tetris game for GNOME
 * Copyright (C) 1999, 2000, 2001 Sebastien Nicoud <seb@eseb.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/

#include "common.h"
#include "csettings.h"
#include "chof.h"
#include "cwhof.h"

// statics

gboolean CWHof::_viewed = false;

// constructors & operators ///////////////////

CWHof::CWHof()
{
	if (debug) g_print("CWHof::CWHof()\n");
	for (gint i=0; i<MAX_NHOFL+2; i++) _data[i] = NULL;
	_game = NULL;
	_state = state_VIEW_ONLY;
	_nhofl = 0;
	_game_index = -1;
	_show_old_score = TRUE;
	_nhofl_to_display = DFLT_NHOFL;
	_already_recorded_game_index = 0;
	_slist = NULL;
	_table = NULL;
	_wh = -1;
}

CWHof::~CWHof()
{
	if (debug) g_print("CWHof::~CWHof()\n");
	free_memory();
}

ostream &operator<< (ostream &os, CWHof &ind)
{
//	if (debug > 2) g_print("CWHof::operator<<\n");
	os	<< "HOF: nhofl=" << ind._nhofl
		<< ", displayed=" << ind._nhofl_to_display
		<< ", old scores:" << ind._show_old_score << endl;
	if (ind._game != NULL) os  << "  game: " << *ind._game;
	else os << "  game: <empty record>" << endl;
	for (gint i=0; i<ind._nhofl; i++) {
		if (ind._data[i] != NULL) os << "  [" << i << "]:" << *ind._data[i];
		else os << "  [" << i << "]: <empty record>" << endl;
	}
	return os;
}

// methods ////////////////////////////

/* append_game_to_hof
 * returns TRUE: ok
 *         FALSE: game already recorded
*/
gboolean CWHof::append_game_to_hof(void)
{
	if (debug) g_print("CWHof::append_game_to_hof\n");
	// read game
	if (_game == NULL) _game = g_new0(CHof,1);
	*_game = *mds;  // copy & cast CGame into CHof
	_game->compute_date();
	// test for presence in list
	for (gint i=0; i<_nhofl; i++) {
		if (_data[i] != NULL && _data[i]->_id == _game->_id) {
			_already_recorded_game_index = i;
			return FALSE;
	}	}
	// append
	if (debug) g_print("  append game at %d\n",_nhofl);
	if (_data[_nhofl] == NULL) _data[_nhofl] = g_new0(CHof,1);
	*_data[_nhofl] = *_game; // copy
	_game_index = _nhofl;
	_nhofl++;
	return TRUE;
}

void CWHof::close_hof(void)
{
	if (debug) g_print("CWHof::close_hof\n");
	if (!_viewed) return;
	if (_slist != NULL) {
		_slist = g_slist_nth(_slist, 0); // rewind
		if (debug > 1) g_print("  free string memory...");
	}
	while (_slist != NULL) {
		if (_slist->data != NULL) {
			g_free(_slist->data);
			_slist->data = NULL;
		}
		_slist = g_slist_next(_slist);
	}
	g_slist_free(_slist); // _slist = NULL
	if (_table != NULL) gtk_clist_clear(GTK_CLIST(_table));
	if (debug > 1) g_print(" memory freed\n");
	// close dialog
	_wh = _window->allocation.height; // keep height setting in the same session
	_viewed = false;
	if (debug) g_print("  close_hof() ends\n");
}

/* does_player_enter_hof
 * returns : 0: no
 *           1: yes, but hidden
 *			 2: yes, top score !
*/
gint CWHof::does_player_enter_hof(void)
{
	if (debug) g_print("CWHof::does_player_enter_hof\n");
	if (_game_index >= MAX_NHOFL) return 0;
	if (_game_index < _nhofl_to_display) return 2;
	return 1;
}

/* free_memory
 * called by the destructor
*/
void CWHof::free_memory(void)
{
	if (debug>1) g_print("CWHof::free_memory\n");

	// call close_hof() just in case the main app is destroyed
	// with the HOF still open
	if (_viewed) close_hof();
	
	for (gint i=0; i<MAX_NHOFL+2; i++) if (_data[i] != NULL) {
		g_free(_data[i]);
		_data[i] = NULL;
	}
	if (_game != NULL) {
		g_free(_game);
		_game = NULL;
	}
}

void CWHof::hof(gboolean update, gint show_old_score, gint nhofl_to_display)
{
	if (debug) g_print(
		"CWHof::show_hof (update=%d, old scores=%d, nhofl to display=%d\n",
		update, show_old_score, nhofl_to_display);

	// init
	_state = state_VIEW_ONLY;
	if (show_old_score == -1) _show_old_score = settings->_show_old_score;
	else {
		if (show_old_score == 0) _show_old_score = FALSE;
		else _show_old_score = TRUE;
	}
	if (nhofl_to_display == -1) _nhofl_to_display = settings->_nhofl;
	else _nhofl_to_display = nhofl_to_display;
	if (_nhofl_to_display > MAX_NHOFL) _nhofl_to_display = MAX_NHOFL;
	else if (_nhofl_to_display < MIN_NHOFL) _nhofl_to_display = MIN_NHOFL;

	// update hof
  	if (read_hof()) return;
	if (update) {
		if (!append_game_to_hof()) {
			// game id is already in the existing hof
			// => don't change anything in the hof, view only
			sort_hof();
			_state = state_ALREADY_RECORDED;
		} else {
			sort_hof();
			gint iret = does_player_enter_hof();
			if (iret == 2) {
				// yes, in top scores
				_state = state_NEW_ENTRY;
				window_name_entry();
				strcpy(_data[_game_index]->_name,_game->_name);
				if (write_hof()) return;
			} else if (iret == 1) {
				// yes, but not shown
				_state = state_NEW_HIDDEN_ENTRY;
				if (write_hof()) return;
			} else {
				// no, but show the line once
				_state = state_NOT_GOOD_ENOUGH;
				strcpy(_data[_game_index]->_name,"YOU ARE HERE");
			}
		}
	}
	if (debug) {
		g_print("  HOF state = ");
		if (_state == state_VIEW_ONLY) g_print("VIEW_ONLY\n");
		else if (_state == state_NOT_GOOD_ENOUGH) g_print("NOT_GOOD_ENOUGH\n");
		else if (_state == state_NEW_ENTRY) g_print("NEW_ENTRY\n");
		else if (_state == state_NEW_HIDDEN_ENTRY) g_print("NEW_HIDDEN_ENTRY\n");
		else if (_state == state_ALREADY_RECORDED) g_print("ALREADY_RECORDED\n");
		else g_print("*** error: undefined state\n");
	}
	// display window
	window_hof();
	if (debug > 2) cout << *this;
	if (debug) g_print("  show_hof() ending\n");
}

/* sort_table
*/
void CWHof::sort_hof(void) {
	if (debug) g_print("CWHof::sort_hof _nhofl=%d\n",_nhofl);

	// compute sort-table
	gint sort[MAX_NHOFL+2];
	gint i;
	gint j;
	for (i=0; i<_nhofl; i++) sort[i] = -1;
	
	for (i=0; i<_nhofl; i++) {
		if (i == 0) {
			// put the first one
			sort[0] = 0;
		} else {
			for (j=0; j<=i; j++) {
				if (sort[j] == -1
				 || _data[i]->is_better_than(_data[sort[j]])) {
					// move sort table down
					for (gint k=_nhofl; k>j; k--) sort[k] = sort[k-1];
					// insert current
					sort[j] = i;
					break; // exit j loop
	}	}	}	}

	// get player rank
	if (_game_index != -1) for (gint i=0; i<_nhofl; i++)
		if (sort[i] == _game_index) { _game_index=i; break; }

	// sort hof using sort-table
	CHof *tmphof[MAX_NHOFL+2];
	for (gint i=0;i<_nhofl;i++) tmphof[i] = _data[i];
	for (gint i=0;i<_nhofl;i++) _data[i] = tmphof[sort[i]];
	return;
}

/* table_add_row
 * input type = 1(new),2(old) otherwise regular
 */
void CWHof::table_add_row(gint rank, gint index)
{
	if (debug>2) g_print("CWHof::table_add_row rank %d (at %d)\n", rank,index);
	if (!_data[rank]->is_version_current()
	 && !_show_old_score) return;
	if (index == -1) index = rank;
	_data[rank]->_rank = rank+1;
	gchar *row[_nc];
	row[0] = _data[rank]->stringify_rank();
	_slist = g_slist_append(_slist, row[0]);
	row[1] = _data[rank]->stringify_score();
	_slist = g_slist_append(_slist, row[1]);
	row[2] = _data[rank]->stringify_difficulty();
	_slist = g_slist_append(_slist, row[2]);
	row[3] = _data[rank]->stringify_level();
	_slist = g_slist_append(_slist, row[3]);
	row[4] = _data[rank]->stringify_layers();
	_slist = g_slist_append(_slist, row[4]);
	row[5] = _data[rank]->stringify_depth();
	_slist = g_slist_append(_slist, row[5]);
	row[6] = _data[rank]->stringify_width();
	_slist = g_slist_append(_slist, row[6]);
	row[7] = _data[rank]->stringify_height();
	_slist = g_slist_append(_slist, row[7]);
	row[8] = _data[rank]->stringify_timer();
	_slist = g_slist_append(_slist, row[8]);
	row[9] = _data[rank]->stringify_name();
	_slist = g_slist_append(_slist, row[9]);
	row[10] = _data[rank]->stringify_date();
	_slist = g_slist_append(_slist, row[10]);
	gtk_clist_append(GTK_CLIST(_table), row);
	gtk_clist_set_selectable(GTK_CLIST(_table), index, false);
}

void CWHof::window_hof(void)
{
	if (debug) g_print("CWHof::window_hof\n");

	GtkWidget *box1, *box2;
	GtkWidget *lbl;
	GtkObject *adj;
	GtkWidget *sbar;
	GtkWidget *sep;
	GtkWidget *button;
	gchar *s;
	gint nb_ranks;
	
	// init	window	
	if (_viewed) {
		if (debug) g_print("CWHof::window_hof: window already open, closing\n");
		//close_hof();
		gtk_widget_destroy(_window); // will emit "destroy" on _window
	}
	_viewed = true;
	
	// create the hof window
	_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	
	gtk_window_set_position(GTK_WINDOW(_window),GTK_WIN_POS_CENTER);
	gtk_window_set_policy(GTK_WINDOW(_window),
		TRUE/*shrink*/, TRUE/*grow*/, TRUE/*auto shrink*/);
	gtk_container_border_width(GTK_CONTAINER(_window), 10);
		
	// window handlers
	gtk_signal_connect(GTK_OBJECT(_window), "destroy",
		GTK_SIGNAL_FUNC(cb_window_hof_destroy), this);

	// main container
	box1 = gtk_vbox_new(FALSE,5);
	gtk_container_add(GTK_CONTAINER(_window),box1);
	gtk_widget_show(box1);
	
	nb_ranks = _nhofl;
	if (nb_ranks > _nhofl_to_display) nb_ranks = _nhofl_to_display;
	
	// congrats label
	if (_state == state_NOT_GOOD_ENOUGH) {
		lbl = gtk_label_new(_(
			"Congrats! You played well.\nUnfortunately, you can't enter the Hall Of Fame :-("));
	} else if (_state == state_NEW_HIDDEN_ENTRY) {
		s = g_strdup_printf(_(
			"Congrats! You played well.\nUnfortunately, you can't enter the TOP %d\nYour rank is: %d"),_nhofl_to_display, _game_index+1);
		lbl = gtk_label_new(s);
		g_free(s);
	} else if (_state == state_NEW_ENTRY) {
		if (_game_index == 0) s = g_strdup(_("You are the best !"));
		else s = g_strdup_printf(_("You are ranked: %d"),_game_index+1);
		lbl = gtk_label_new(s);
		g_free(s);
	} else if (_state == state_ALREADY_RECORDED) {
		gchar *sd = _data[_already_recorded_game_index]->stringify_date();
		gchar *sn = _data[_already_recorded_game_index]->stringify_name();
		s = g_strdup_printf(_(
			"Congrats! You played well but...\nthe game you just completed has already been recorded on\n%s by %s"),sd,sn);
		lbl = gtk_label_new(s);
		g_free(s);
		g_free(sd);
		g_free(sn);
	} else if (nb_ranks == 0) {
		// VIEW ONLY - the HOF is empty
		lbl = gtk_label_new(_("No one is famous, yet..."));
	} else {
		// VIEW ONLY
		s = g_strdup_printf(_("Gnome 3D Tetris TOP %d"),_nhofl_to_display);
		lbl = gtk_label_new(s);
		g_free(s);
	}
	gtk_box_pack_start(GTK_BOX(box1),lbl,FALSE,FALSE,0);
	gtk_widget_show(lbl);

	if (debug) {
		g_print(
			"  nb_ranks=%d, nhofl=%d, nhofl_to_display=%d\n",
			nb_ranks, _nhofl, _nhofl_to_display);
	}
	
	if (nb_ranks) {
		// h box
		box2 = gtk_hbox_new(FALSE,0);
		gtk_box_pack_start(GTK_BOX(box1),box2,TRUE,TRUE,0);
		gtk_widget_show(box2);

		// create the scroollbar
		adj = gtk_adjustment_new(0,0,1000,1,10,0);
		sbar = gtk_vscrollbar_new(GTK_ADJUSTMENT(adj));
		gtk_box_pack_start(GTK_BOX(box2),sbar,FALSE,FALSE,0);
		gtk_widget_show(sbar);

		gchar *titles[] = {
			"Rank", "Score", "Difficulty", "Level", "Layers",
			"Depth", "Width", "Height", "Time", "Name", "Date" };
		_table = gtk_clist_new_with_titles(_nc, titles);
		gtk_box_pack_start(GTK_BOX(box2),_table,FALSE,FALSE,0);
		gtk_widget_show(_table);
		gtk_clist_set_selection_mode(GTK_CLIST(_table),GTK_SELECTION_SINGLE);
		gtk_clist_column_titles_passive(GTK_CLIST(_table));
		for (gint i=0; i<10; i++)
			gtk_clist_set_column_justification(GTK_CLIST(_table),
				i, GTK_JUSTIFY_CENTER);
		gtk_clist_set_column_justification(GTK_CLIST(_table),
			10, GTK_JUSTIFY_LEFT);
		gtk_clist_set_vadjustment(GTK_CLIST(_table),GTK_ADJUSTMENT(adj));

		// fill out list
		gint irank = 0;
		for (irank=0; irank<nb_ranks; irank++) table_add_row(irank);

		// one extra line
		gint game_index_in_clist = 0;
		if (_state == state_NOT_GOOD_ENOUGH
		 || _state == state_NEW_HIDDEN_ENTRY) {
		 	nb_ranks++;
			table_add_row(_game_index, irank);
			game_index_in_clist = irank;
		} else if (_state == state_NEW_ENTRY)
			game_index_in_clist = _game_index;

		// disable column resizing
		gtk_clist_columns_autosize(GTK_CLIST(_table));
		for (gint i=0; i<_nc; i++)
			gtk_clist_set_column_resizeable(GTK_CLIST(_table), i, false);
	
		// highlight the new entry if there is
		if (_state == state_NEW_ENTRY
		 || _state == state_NOT_GOOD_ENOUGH
		 || _state == state_NEW_HIDDEN_ENTRY) {
		 	GdkColor c1;
			c1.red = 0x0000;
			c1.green = 0x0000;
			c1.blue = 0x0000;
		 	GdkColor c2;
			c2.red = 0xFFFF;
			c2.green = 0x4444;
			c2.blue = 0x0000;
		 	gtk_clist_set_background(GTK_CLIST(_table),
				game_index_in_clist, &c1);
		 	gtk_clist_set_foreground(GTK_CLIST(_table),
				game_index_in_clist, &c2);	
			gtk_clist_select_row(GTK_CLIST(_table),
				game_index_in_clist, 0);
		}
	} // if (nb_rank)

	// sep
	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(box1),sep,FALSE,FALSE,5);
	gtk_widget_show(sep);
	
	// close button
	box2 = gtk_hbox_new(FALSE,5);
	gtk_box_pack_start(GTK_BOX(box1),box2,FALSE,FALSE,0);
	gtk_widget_show(box2);
	button = gnome_stock_button(GNOME_STOCK_BUTTON_CLOSE);
	gtk_box_pack_end(GTK_BOX(box2),button, FALSE, FALSE,0);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(_window));
		// sends "destroy" caught by _window, as defined above
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_draw_default(button);
	gtk_widget_show(button);
		
	// resize if too big
	gtk_widget_realize(_window);
	gint w = _window->allocation.width;
	gint h = _window->allocation.height;
	if (debug) g_print("  hof is %d x %d\n",w,h);
	if (_wh != -1 && h > _wh && nb_ranks>3) h = _wh;
	if (h > gdk_screen_height()-160) h = gdk_screen_height()-160;
	if (debug) g_print("  screen height=%d, saved height=%d -resized to %d\n",
		gdk_screen_height(), _wh, h);
	gtk_widget_set_usize(GTK_WIDGET(_window), w, h);

	// show the whole damn thing
	gtk_widget_show_all(_window);
}

void CWHof::window_name_entry(void)
{
	if (debug) g_print("CWHof::window_name_entry\n");
	GtkWidget *window_entry;
	GtkWidget *lbl;
	GtkWidget *entry;
	GtkWidget *box;

	window_entry = gnome_dialog_new (_("Gnome 3D Tetris Hall Of Fame"),
		GNOME_STOCK_BUTTON_OK, NULL);
	gtk_window_set_position(GTK_WINDOW(window_entry),GTK_WIN_POS_CENTER);

	// window handlers
	gtk_signal_connect_object(GTK_OBJECT(window_entry), "destroy",
		GTK_SIGNAL_FUNC(cb_window_name_entry_destroy), GTK_OBJECT(window_entry));

	// stuff
	box = GNOME_DIALOG(window_entry)->vbox;
	lbl = gtk_label_new(_("Congratulation !\n\
You're entering the Gnome 3D Tetris Hall Of Fame.\n\
Please enter your name\n"));
	gtk_box_pack_start(GTK_BOX(box), lbl, FALSE, FALSE, 0);
	entry = gtk_entry_new_with_max_length(MAX_NAME_LEN);
	if (_game->_name != NULL)
		gtk_entry_set_text(GTK_ENTRY(entry),_game->_name);
	gtk_box_pack_start(GTK_BOX(box), entry, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(entry), "changed",
		GTK_SIGNAL_FUNC(cb_entry_name_changed), this);

	gnome_dialog_button_connect_object(GNOME_DIALOG(window_entry),
		0/*close*/, GTK_SIGNAL_FUNC(cb_window_name_entry_destroy),
		GTK_OBJECT(window_entry));
	gtk_signal_connect(GTK_OBJECT(window_entry), "key_press_event",
		GTK_SIGNAL_FUNC(cb_window_name_entry_key_press_event),NULL);

	gtk_widget_show_all(window_entry);
	gnome_dialog_run_and_close (GNOME_DIALOG (window_entry));
	if (debug) g_print("  window_name_entry, name = >%s<\n",_game->_name);
}

// I/O stuff //////////////////////////////////

/* read_hof
 *  reads the hof file and updates the hof data structure
 *  returns: 0 -> ok
 *           1 -> no file
 *           - -> error occured
*/
guint CWHof::read_hof(void)
{
	if (debug) g_print("WCHof::read_hof\n");
	
	// open hof file
	ifstream fin;
	gchar *filename = g_strdup(GNO3DTET_HOFABSFILE);
	fin.open(filename);
	if (!fin.is_open()) {
		if (debug) g_print("  opening >%s< failed\n",filename);
		// try current dir
		g_free(filename);
		filename = g_strdup(HOFFILE);
		fin.open(filename);
		if (!fin.is_open()) {
			if (debug) g_print("  opening >%s< failed\n",filename);
			// try home dir
			g_free(filename);
			filename = g_strdup_printf("%s%s%s",
				g_get_home_dir(), G_DIR_SEPARATOR_S, HOFFILE);
			fin.open(filename);
			if (!fin.is_open()) {
				if (debug) g_print("  opening >%s< failed\n",filename);
				g_free(filename);
				if (debug) g_print("  read_hof exit elvl 1: can't open file\n");
				return 1;
	}	}	}
	fin.close();
	if (debug) g_print("  read_hof: reading file: >%s<...\n",filename);

	_game_index = -1;
	_nhofl = 0;
	CHof temp;
	guint iret = 0;
	iret = temp.read_game(filename, TRUE/*from the beginning*/);
	while (iret == 0/*ok*/) {
		if (_data[_nhofl] == NULL) _data[_nhofl] = g_new0(CHof,1);
		*_data[_nhofl] = temp;
		_nhofl++;
		if (_nhofl >= MAX_NHOFL) break; // limit reached, exit while(iret)
		iret = temp.read_game(filename, FALSE/*at current pos*/);
	}
	if (iret == 1/*EOF, it's Ok*/) iret = 0;
	if (debug) g_print("  read hof: nb records read = %d, exit elvl %d\n",
		_nhofl,iret);
	g_free(filename);
	return iret;
}

/* write_hof
 *  saves the hof data structure in the hof file
 *  returns: 0 -> ok
 *           1 -> can't open file
 *           - -> error occured
*/
gint CWHof::write_hof(void) {

	if (debug) g_print("CWHof::write_hof\n");
	if (_nhofl == 0) return 0; // table is empty => nothing to save
	if (_state == state_VIEW_ONLY
	 || _state == state_ALREADY_RECORDED
	 || _state == state_NOT_GOOD_ENOUGH)
	 	return 0; // nothing changed => nothing to save
	
	gint iret = 0;
	gint nb_records = _nhofl;
	
	// truncate table if too big
	if (nb_records > MAX_NHOFL) nb_records = MAX_NHOFL;

	// look for file name & test
	ofstream fout;
	gchar *filename = g_strdup(GNO3DTET_HOFABSFILE);
	fout.open(filename);
	if (!fout.is_open()) {
		if (debug) g_print("  opening >%s< failed\n",filename);
		// try current dir
		filename = g_strdup(HOFFILE);
		fout.open(filename);
		if (!fout.is_open()) {
			if (debug) g_print("  opening >%s< failed\n",filename);
			// try home dir
			filename = g_strdup_printf("%s%s%s",
				g_get_home_dir(), G_DIR_SEPARATOR_S, HOFFILE);
			fout.open(filename);
			if (!fout.is_open()) {
				if (debug) g_print("  opening >%s< failed\n",filename);
				g_free(filename);
				return 1;
	}	}	}
	fout.close();
	if (debug) g_print("  write_hof: writing to >%s<...\n",filename);
	for (gint i=0; i<nb_records; i++) {
			iret = _data[i]->write_game(filename, i==0? TRUE:FALSE);
			if (iret) break; // halt writing
	}
	if (debug) g_print("  write_hof, exit elvl %d\n",iret);
	g_free(filename);
	return iret;
}

// Callbacks /////////////////////////////////////////////////////////////////

/* cb_close_hof
 * closes the main window
 * called when destroyed or closed
*/
void CWHof::cb_window_hof_destroy(GtkObject *widget, gpointer data)
{
	if (debug > DBG_CB) g_print("CWHof::cb_window_hof_destroy\n");
	((CWHof *)data)->close_hof();
}

/* cb_entry_name_changed
 *  called when user type in the name entry widget
 *  updates the g_hofl global variable
*/
void CWHof::cb_entry_name_changed(GtkObject *widget, gpointer data)
{
	if (debug > DBG_CB) g_print("CWHof::cb_entry_name_changed\n");
	strncpy(((CWHof *)data)->_game->_name,
		gtk_entry_get_text(GTK_ENTRY(widget)),
		MAX_NAME_LEN);
}

/* cb_window_name_entry_destroy
 * closes window
*/
void CWHof::cb_window_name_entry_destroy(GtkObject *widget, gpointer data)
{
	if (debug > DBG_CB) g_print("CWHof::cb_window_name_entry_destroy\n");
	gnome_dialog_close(GNOME_DIALOG(widget));
}

/* cb_window_name_entry_key_press_event
 * closes the name entry dialog w when return key pressed
*/
void CWHof::cb_window_name_entry_key_press_event(GtkWidget *widget,
	GdkEvent *event, gpointer data)
{
	if (debug > DBG_CB)
		g_print("CWHof::cb_window_name_entry_key_press_event\n");
	switch ((gint)(((GdkEventKey *)event)->keyval)) {
		case GDK_Return: case GDK_KP_Enter:
		if (debug > DBG_CB) g_print("  'Return' key pressed\n");
		gnome_dialog_close(GNOME_DIALOG(widget));
		break;
		default: // any other key // nothing to do
		break;
}	}
