/* gno3dtet: 3D Tetris game for GNOME
 * Copyright (C) 1999, 2000, 2001 Sebastien Nicoud <seb@eseb.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/

#include "common.h"
#include "csound.h"
#include "cpiece.h"
#include "csettings.h"
#include "cgame.h"
#include "cwsettings.h"
#include "cwhof.h"
#include "cwmain.h"
#include "main.h"

// common globals /////////////////////////////////////////////////////////////

gint			debug = 0; // debug level
CSettings		*settings = NULL; // official settings
CWSettings		*wsets = NULL; // preferences window
CWMain			*mwin = NULL; // main window
CGame			*mds = NULL; // main data structure
CWHof			*whof = NULL; // HOF window
CSound			sounds[NUMBER_SOUND_EVENT+1]; // sounds db

// main.cc globals ////////////////////////////////////////////////////////////

guint 			keys[16]; // key binding string

// command line arguments /////////////////////////////////////////////////////

static gint		g_opt_debug = 0;
static gint		g_opt_timer = 0;
static gint 	g_opt_xposition    = 0;
static gint 	g_opt_yposition    = 0;
static gint 	g_opt_windowwidth  = 0;
static gint 	g_opt_windowheight = 0;
static gint		g_opt_nhofl  = 0;
static gint 	g_opt_diff   = 0;
static gint 	g_opt_speed  = 0;
static gint 	g_opt_width  = 0;
static gint 	g_opt_height = 0;
static gint 	g_opt_depth  = 0;
static gboolean	g_opt_show_toolbar = TRUE;
static gboolean	g_opt_sound_support = TRUE;

static const struct poptOption options[] = {
	{"debug", 'd', POPT_ARG_INT, &g_opt_debug, 0,
	 N_("Debugging mode level [0..3]"), N_("INT")},
	{"timer", 't', POPT_ARG_INT, &g_opt_timer, 0,
	 N_("Delay between 2 frames, in msec. [1..100]"), N_("INT")},
	{"xpos", 'x', POPT_ARG_INT, &g_opt_xposition, 0,
	 N_("X position of window, in pixels"), N_("INT")},
	{"ypos", 'y', POPT_ARG_INT, &g_opt_yposition, 0,
	 N_("Y position of window, in pixels"), N_("INT")},
	{"windowwidth", 'i', POPT_ARG_INT, &g_opt_windowwidth, 0,
	 N_("Window width, in pixels"), N_("INT")},
	{"windowheight", 'j', POPT_ARG_INT, &g_opt_windowheight, 0,
	 N_("Window height, in pixels"), N_("INT")},
	{"difficulty", 'p', POPT_ARG_INT, &g_opt_diff, 0,
	 N_("Piece difficulty [1..3]"), N_("INT")},
	{"speed", 's', POPT_ARG_INT, &g_opt_diff, 0,
	 N_("Piece moving speed [1..3]"), N_("INT")},
	{"width", 'w', POPT_ARG_INT, &g_opt_width, 0,
	 N_("World width, in nb of cubes"), N_("INT")},
	{"height", 'h', POPT_ARG_INT, &g_opt_height, 0,
	 N_("World height, in nb of cubes"), N_("INT")},
	{"depth", 'd', POPT_ARG_INT, &g_opt_depth, 0,
	 N_("World depth, in nb of cubes"), N_("INT")},
	{"hoflines", 'f', POPT_ARG_INT, &g_opt_nhofl, 0,
	 N_("Number of lines in the Hall Of Fame"), N_("INT")},
	{NULL, '\0', 0, NULL, 0}
};

// callbacks //////////////////////////////////////////////////////////////////

/* client_die
 * Session Management: die
*/
static gint client_die (GnomeClient *client, gpointer client_data)
{
	if (debug > DBG_CB) g_print("cb: client dies\n");
	gtk_main_quit();
	return FALSE;
}    

/* client_save_yourself
 * Session Management: save_yourself
 * saves game state
*/
static gint client_save_yourself(
	GnomeClient *client,
	gint phase,
	GnomeSaveStyle save_style,
	gint shutdown, 
	GnomeInteractStyle interact_style,
	gint fast,
	gpointer client_data)
{
	if (debug > DBG_CB) g_print ("cb main> client save yourself\n");

	gchar *prefix= gnome_client_get_config_prefix (client);
	gchar *argv[]= { "rm", "-r", NULL };
  
	gchar *filename = g_strdup_printf("%s%s%s",
		g_get_home_dir(), G_DIR_SEPARATOR_S, SAVED_SESSION);
	gint iret = 0;//	gint iret = game.write_game(filename); // FIXME
	
	argv[2]= gnome_config_get_real_path (prefix);
	gnome_client_set_discard_command (client, 3, argv);

	if (debug > DBG_CB) g_print ("cb main< client save yourself, \
		saved in >%s<, exited elvl %d\n",filename,iret);
	g_free(filename);
	return TRUE;
}

// functions //////////////////////////////////////////////////////////////////

/* setup_settings
 * setup the global setting values (settings) and the debug level
 * get input from:
 * 	1. command line
 *	2. saved config
 *  3. defaults
 * this function is called once, at the very beginning
 * G: settings
*/
void setup_settings(void)
{
    gchar config_str[256];

	// debug level
	if (g_opt_debug <= 0) debug = 0;
	else if (g_opt_debug > 3) debug = 3;
	else debug = g_opt_debug;
	if (debug) g_print("main::setup_settings, debug level = %d\n",debug);

	// timer
	if (g_opt_timer == 0) g_opt_timer = MAIN_TIMER;
	else if (g_opt_timer < 1) g_opt_timer = 1;
	else if (g_opt_debug > 100) g_opt_timer = 100;

	// window width and height
	if (g_opt_windowwidth == 0)	{
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_WINDOWWIDTH, DFLT_WINDOWWIDTH);
		g_opt_windowwidth = gnome_config_get_int_with_default(config_str, NULL);
	}
	if (g_opt_windowwidth < 100 || g_opt_windowwidth > gdk_screen_width())
		g_opt_windowwidth = DFLT_WINDOWWIDTH;

	if (g_opt_windowheight == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_WINDOWHEIGHT, DFLT_WINDOWHEIGHT);
		g_opt_windowheight = gnome_config_get_int_with_default(config_str, NULL);
	}
	if (g_opt_windowheight < 100 || g_opt_windowheight > gdk_screen_height())
		g_opt_windowheight = DFLT_WINDOWHEIGHT;

	// nb lines in hof
	if (g_opt_nhofl == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_NHOFL, DFLT_NHOFL);
		settings->_nhofl = gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_nhofl = g_opt_nhofl;
	if (settings->_nhofl < MIN_NHOFL) settings->_nhofl = MIN_NHOFL;
	else if (settings->_nhofl > MAX_NHOFL) settings->_nhofl = MAX_NHOFL;
	
	// show old scores
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_OLD_SCORE);
	settings->_show_old_score =
		gnome_config_get_bool_with_default(config_str, NULL);

	// world width
	if (g_opt_width == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_WIDTH, DEF_WORLD_WIDTH);
		settings->_w = gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_w = g_opt_width;
	if (settings->_w < MIN_WORLD_WIDTH) settings->_w = MIN_WORLD_WIDTH;
	else if (settings->_w > MAX_WORLD_WIDTH) settings->_w = MAX_WORLD_WIDTH;

	// world height
	if (g_opt_height == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_HEIGHT, DEF_WORLD_HEIGHT);
		settings->_h = gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_h = g_opt_height;
	if (settings->_h < MIN_WORLD_HEIGHT) settings->_h = MIN_WORLD_HEIGHT;
	else if (settings->_h > MAX_WORLD_HEIGHT) settings->_h = MAX_WORLD_HEIGHT;

	// world depth
	if (g_opt_depth == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_DEPTH, DEF_WORLD_DEPTH);
		settings->_d = gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_d = g_opt_depth;
	if (settings->_d < MIN_WORLD_DEPTH) settings->_d = MIN_WORLD_DEPTH;
	else if (settings->_d > MAX_WORLD_DEPTH) settings->_d = MAX_WORLD_DEPTH;

	// difficulty
	if (g_opt_diff == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_DIFF, DEF_DIFFICULTY);
		settings->_difficulty =
			gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_difficulty = g_opt_diff;
	if (settings->_difficulty <= CSettings::diff_EASIEST_EXCLUDED)
		settings->_difficulty = CSettings::diff_EASIEST_EXCLUDED+1;
	else if (settings->_difficulty >= CSettings::diff_HARDEST_EXCLUDED)
		settings->_difficulty = CSettings::diff_HARDEST_EXCLUDED-1;

	// speed
	if (g_opt_speed == 0) {
		sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
			CONFIG_SPEED, DEF_SPEED);
		settings->_speed =
			gnome_config_get_int_with_default(config_str, NULL);
	} else settings->_speed = g_opt_diff;
	if (settings->_speed <= CSettings::speed_SLOWEST_EXCLUDED)
		settings->_speed = CSettings::speed_SLOWEST_EXCLUDED+1;
	else if (settings->_speed >= CSettings::speed_FASTEST_EXCLUDED)
		settings->_speed = CSettings::speed_FASTEST_EXCLUDED-1;

	// show toolbar
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_TOOLBAR);
	g_opt_show_toolbar = gnome_config_get_bool_with_default(config_str, NULL);

	// show next piece
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_NEXT_PIECE);
	settings->_show_np =
		gnome_config_get_bool_with_default(config_str, NULL);

	// show scale
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_SCALE);
	settings->_show_scale =
		gnome_config_get_bool_with_default(config_str, NULL);

	// show score
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_SCORE);
	settings->_show_score =
		gnome_config_get_bool_with_default(config_str, NULL);

	// show stats
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SHOW_STATS);
	settings->_show_stats =
		gnome_config_get_bool_with_default(config_str, NULL);

	// key binding
	gchar *dflt_key_string = DFLT_KEY_BINDING;

	// x axis keys
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_XNEG, dflt_key_string[key_XNEG]);
	keys[key_XNEG] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_XPOS, dflt_key_string[key_XPOS]);
	keys[key_XPOS] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// y axis keys
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_YNEG, dflt_key_string[key_YNEG]);
	keys[key_YNEG] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_YPOS, dflt_key_string[key_YPOS]);
	keys[key_YPOS] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// z axis keys
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_ZNEG, dflt_key_string[key_ZNEG]);
	keys[key_ZNEG] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_ZPOS, dflt_key_string[key_ZPOS]);
	keys[key_ZPOS] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// transl keys
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_LEFT, dflt_key_string[key_LEFT]);
	keys[key_LEFT] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_RIGHT, dflt_key_string[key_RIGHT]);
	keys[key_RIGHT] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_DOWN, dflt_key_string[key_DOWN]);
	keys[key_DOWN] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_UP, dflt_key_string[key_UP]);
	keys[key_UP] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// drop
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_DROP, dflt_key_string[key_DROP]);
	keys[key_DROP] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// pause & toggle next piece
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_PAUSE, dflt_key_string[key_PAUSE]);
	keys[key_PAUSE] = (guint)gnome_config_get_int_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=%d", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_KEY_TOGGLENP, dflt_key_string[key_TOGGLENP]);
	keys[key_TOGGLENP] = (guint)gnome_config_get_int_with_default(config_str, NULL);

	// focus behavior
	sprintf(config_str,"%s/%s/%s=false", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_PAUSE_LOSTFOCUS);
	settings->_pause_on_nofocus =
		gnome_config_get_bool_with_default(config_str, NULL);
	sprintf(config_str,"%s/%s/%s=false", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_RESUME_FOCUS);
	settings->_resume_on_focus =
		gnome_config_get_bool_with_default(config_str, NULL);
	
	// setup key bindings string
	for (gint i=0; i<key_NONE; i++) settings->_keys[i] = keys[i];

	if (debug) {
		cout << "  " << *settings;
		cout << "  toolbar=" << g_opt_show_toolbar;
		cout << ", timer=" << g_opt_timer << endl;
		cout << "  default key binding=>" << dflt_key_string << "<";
		gchar mapped_keys[key_NONE];
		for (gint i=0; i<key_NONE; i++) mapped_keys[i] = (gchar)keys[i];
		mapped_keys[key_NONE] = 0;
		cout << ", customized key binding=>" << mapped_keys << "<";
		cout << endl;
	}	

	// sound support
	sprintf(config_str,"%s/%s/%s=true", GNO3DTET_EXENAME, CONFIG_PREFERENCES,
		CONFIG_SOUNDSUPPORT);
	settings->_sound_support =
		gnome_config_get_bool_with_default(config_str, NULL);

	return;
}

#ifndef GNOME_CLIENT_RESTARTED
#define GNOME_CLIENT_RESTARTED(client) \
(GNOME_CLIENT_CONNECTED (client) && \
 (gnome_client_get_previous_id (client) != NULL) && \
 (strcmp (gnome_client_get_id (client), \
  gnome_client_get_previous_id (client)) == 0))
#endif /* GNOME_CLIENT_RESTARTED */

/* main
*/
int main(int argc, char *argv[])
{    
	debug = 0;
	
	// i18n   
	bindtextdomain (PACKAGE, GNOMELOCALEDIR);
	textdomain (PACKAGE);

	// init
	gnome_init_with_popt_table (GNO3DTET_EXENAME, GNO3DTET_VERSION,
		argc, argv, options, 0, NULL);

   	g_print ("%s %s\n", GNO3DTET_EXENAME, GNO3DTET_VERSION);
//	gnome_sound_init("localhost"); // should I do it or not ?
	
	// setup general settings, debug level included
    if (debug) g_print("main: create CSettings class instance\n");
	settings = new CSettings();
	setup_settings();

	if (debug) {
    	gchar arg_str[1024]; arg_str[0] = '\0';
    	gint i=0;
		while (argv[i] != NULL) sprintf(arg_str,"%s %s",
			arg_str,(gchar *)argv[i++]);
    	g_print("call: %s\n",arg_str);
   		if (G_HAVE_GINT64) g_print("main: int64 ok\n");
		else g_print("main: int64 ko\n");
		g_print("main: makefile defs: HOFDIR=%s, HOFFILE=%s, SOUNDDIR=%s\n",HOFDIR,HOFFILE,SOUNDDIR);
	}

	// create instances
    if (debug) g_print("main: create CWMain class instance\n");
	mwin = new CWMain();
	mwin->create_timer(g_opt_timer);
	
	// setup window settings
	mwin->set_window_geometry(
		g_opt_xposition, g_opt_yposition,
		g_opt_windowwidth, g_opt_windowheight);

	// setup toggle items in the Settings menu
	mwin->toggle_toolbar(g_opt_show_toolbar ?
		CWMain::toggle_ON : CWMain::toggle_OFF);
	mwin->toggle_sound_support(CWMain::toggle_UPDATE/*according to settings*/);	
	mwin->toggle_all(CWMain::toggle_UPDATE/*according to settings*/);

	// sounds
	csound_init_sounds();
	if (debug > 1) {
		cout << "sound support = " << settings->_sound_support << endl;
		for (gint i=0; i<NUMBER_SOUND_EVENT; i++)
			cout << sounds[i];
	}

    if (debug) g_print("main: create CGame class instance\n");
	mds = new CGame();
	mds->create_pieces(); // pieces are static

    if (debug) g_print("main: create CWSettings class instance\n");
	wsets = new CWSettings();

    if (debug) g_print("main: create CWHof class instance\n");
	whof = new CWHof();

	// session management
    if (debug) g_print("main: start session mgt\n");
	GnomeClient *client = gnome_master_client();

	// load saved session if needed
	if (GNOME_CLIENT_RESTARTED (client)) {
//		mds->restart();
		mds->init(); // FIXME
	} else {
		mds->init();
	}
	gtk_signal_connect (GTK_OBJECT (client), "save_yourself",
		GTK_SIGNAL_FUNC (client_save_yourself), argv[0]);
    gtk_signal_connect (GTK_OBJECT (client), "die",
		GTK_SIGNAL_FUNC (client_die), NULL);
	
	// start gtk event loop
	if (debug) g_print("main: entering gtk loop\n");	
	gtk_main();
	mwin->save_window_geometry();
	
	// destroy instances
	if (debug) g_print("main: destroying instances\n");
	delete whof;
	delete settings;
	delete wsets;
	delete mwin;
	mds->destroy_pieces();
	delete mds;
	
	if (debug) g_print("main: exiting, elvl 0.\n");
	return 0;
}
